/**
 * The SceneLoader is in charge of loading scenes.
 * If need be, it may use a HTTP request to fetch the scene first.
 * If the scene is already cached, it will simply be loaded from memory.
 */
function SceneLoader() {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;

    // All currently loaded and cached scenes.
    var cachedScenes = {};

    // Error message shown when loading fails.
    var sceneLoadErrorMessage = "Scene loading failed! " +
            "This probably broke something. " +
            "Please reload the page and try again. " +
            "If that doesn't work, try clearing your browser cache (ctrl+F5 on Chrome)." +
            "If the problem persists, please do notify the creator of Kobold Adventure. " +
            "Keep in mind that it may just be your internet connection acting up!";

    // Queue for registration functions.
    var registrationQueue = [];

    // Whether or not registration processing has started.
    var registerStarted = false;

    // How many files advanced loading is currently processing.
    var loadingFiles = 0;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Builds and returns the path to the root of the scene.
     * @param {string} sceneName The name of the scene.
     * @param {string} subfolder The subfolder the scene is contained in. Optional parameter.
     * @returns {string} The path to the root of the scene.
     */
    function buildSceneRoot(sceneName, subfolder) {
        var root = "scenes" + "/";
        if (subfolder) {
            root += subfolder + "/";
        }
        root += sceneName + "/";
        return root;
    }

    /**
     * Builds and returns a path to the main JSON file of a scene.
     * @param {string} sceneName The name of the scene.
     * @param {string} subfolder The subfolder the scene is contained in. Optional parameter.
     * @returns {string} The path to the main JSON file of the scene.
     */
    function buildJsonFilePath(sceneName, subfolder) {
        return buildSceneRoot(sceneName, subfolder) + sceneName + ".json";
    }

    /**
     * Builds and returns a path to a JavaScript file of a scene.
     * @param {string} sceneName The name of the scene.
     * @param {string} fileName The name of the JavaScript file.
     * @param {string} subfolder The subfolder the scene is contained in. Optional parameter.
     * @returns {string} The path to the JavaScript file.
     */
    function buildJSFilePath(sceneName, fileName, subfolder) {
        return buildSceneRoot(sceneName, subfolder) + fileName + ".js";
    }

    /**
     * Function to be called when something fails to load.
     * Shows an error-alert and hides the loading spinner.
     */
    function loadError() {
        global.error(sceneLoadErrorMessage);
    }

    /**
     * Recursive, asynchronous JavaScript loader.
     * Also known as "old loading".
     * Loads the first element of the passed Queue as a JavaScript file from the scene's root directory.
     * After loading, calls this function again with the element removed.
     * If the queue is empty, finishes loading by calling the scene's load function and hides the loading indicator.
     * @param {type} sceneName The name of the scene.
     * @param {array} queue The queue of files to load.
     * @param {string} subfolder The subfolder the scene can be found in. Optional parameter.
     */
    function loadJavaScriptRecursive(sceneName, queue, subfolder) {
        if (!queue.length) {
            cachedScenes[sceneName].load();
            global.ui.loadbar.hide();
            global.ui.resetScroll();
            return;
        }

        global.ui.loadbar.update(queue.length);
        var file = queue[0];

        $.getScript(buildJSFilePath(sceneName, file, subfolder))
                .done(function () {
                    queue.shift();
                    loadJavaScriptRecursive(sceneName, queue, subfolder);
                })
                .fail(loadError);
    }

    /**
     * Loads a single JavaScript file using the advanced loading method.
     * @param {type} sceneName The name of the scene.
     * @param {string} subfolder The subfolder the scene can be found in. Optional parameter.
     * @param {string} fileName The name of the file to load.
     */
    function loadJavaScript(sceneName, subfolder, fileName) {
        return new Promise(function (resolve, reject) {
            var script = document.createElement("script");
            script.type = `text/javascript`;
            script.onerror = reject;
            script.src = buildJSFilePath(sceneName, fileName, subfolder);
            script.onload = function () {
                loadingFiles--;
                global.ui.loadbar.update(loadingFiles);
                resolve();
            };
            document.body.appendChild(script); // Don't use jQuery for this.
        });
    }

    /**
     * Simultaneous, asynchronous JavaScript loader.
     * Also known as "advanced loading".
     * Loads all required files simultaneously, letting the browser manage download multiplicity.
     * @param {type} sceneName The name of the scene.
     * @param {array} queue The queue of files to load.
     * @param {string} subfolder The subfolder the scene can be found in. Optional parameter.
     */
    function loadJavaScriptSimultaneously(sceneName, queue, subfolder) {
        registerStarted = false;
        loadingFiles = queue.length;
        var promises = [];
        for (var i = 0; i < queue.length; i++) {
            promises.push(loadJavaScript(sceneName, subfolder, queue[i]));
        }

        $.when.apply($, promises).then(function () {
            setTimeout(function () {
                that.registerStart();
                cachedScenes[sceneName].load();
                global.ui.loadbar.hide();
                global.ui.resetScroll();
            }, 50);
        }, loadError);
    }

    /**
     * Loads a scene's JSON via http.
     * Calls the recursive JavaScript loading function when done,
     * or the newer simultaneous loading depending on the settings.
     * Shows the loading indicator at the start.
     * The path of the loaded file will be equal to "scenes/subfolder/sceneName/sceneName.json".
     * @param {string} sceneName The name of the scene.
     * @param {string} subfolder The subfolder the scene can be found in. Optional parameter.
     */
    function httpLoad(sceneName, subfolder) {
        $.getJSON(buildJsonFilePath(sceneName, subfolder))
                .done(function (data) {
                    global.ui.loadbar.show(data.length);
                    if (global.settings.getOldLoad()) {
                        loadJavaScriptRecursive(sceneName, data, subfolder);
                    } else {
                        loadJavaScriptSimultaneously(sceneName, data, subfolder);
                    }

                })
                .fail(loadError);
    }

    /**
     * Processes the registration queue, emptying it one entry at a time, calling each registration function.
     * Does nothing if the registration process has not yet started.
     */
    function processRegistrationQueue() {
        if (!registerStarted) {
            return;
        }
        while (registrationQueue.length) {
            registrationQueue.shift()();
        }
    }

    // PUBLIC MEMBERS \\        this.public = "test";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Loads a scene.
     * If cached, will just call the scene's load method.
     * If not cached, will dynamically load the scene using a http call.
     * @param {string} sceneName The name of the scene to load.
     * @param {string} subfolder An optional parameter declaring the subfolder the scene is in. 
     *                           The folder "scenes" and the folder with the name of the scene itself are always implied.
     */
    this.loadScene = function (sceneName, subfolder) {
        if (cachedScenes[sceneName]) {
            cachedScenes[sceneName].load();
        } else {
            httpLoad(sceneName, subfolder);
        }
    };

    /**
     * Registers a scene with this SceneLoader.
     * Should ideally be immediately called from the scene's JavaScript file.
     * @param {string} name The name of the loaded scene. Will be the key to which the scene is registered.
     * @param {Scene} scene An instance of the scene-object, which will be registered with the name as its key.
     */
    this.registerScene = function (name, scene) {
        cachedScenes[name] = scene;
    };


    /**
     * Queues a scene registration, and immediately processes all pending registrations if registration processessing has been started.
     * @param {function} register Registration function.
     */
    this.queueRegister = function (register) {
        registrationQueue.push(register);
        processRegistrationQueue();
    };

    /**
     * Must be called when registration of scenes can begin.
     * The scene registration queue will remain filled until this function has been called.
     */
    this.registerStart = function () {
        registerStarted = true;
        processRegistrationQueue();
    };
}
